module Main exposing (..)

import Browser
import Data.DataLoader exposing (decode)
import Html exposing (Html, a, button, div, i, nav, text)
import Html.Attributes
import Html.Events exposing (onClick)
import Http
import Model exposing (Global, Local, Model, Progress(..), View(..))
import Msg exposing (Msg(..))
import Triple
import Utils exposing (buildErrorMessage)
import Views.BarPlot as BarPlotView
import Views.DataTableView as DataTableView
import Views.ForceDirectedGraph as ForceDirectedGraphView
import Views.LinePlot as LinePlotView
import Views.MapView as MapView



-- MAIN


main : Program () Model Msg
main =
    Browser.element
        { init = init
        , update = update
        , view = view
        , subscriptions = subscriptions
        }



-- MODEL


init : () -> ( Model, Cmd Msg )
init () =
    let
        global : Global
        global =
            { view = Home
            , progress = Loading
            , data = []
            , selectedCountry = Nothing
            , selectedConsulateCountry = Nothing
            }

        local : Local
        local =
            { map = MapView.init |> Tuple.first
            , dataTable = DataTableView.init
            , forcedirectedgraph = ForceDirectedGraphView.firstInit
            , barplot = BarPlotView.init
            , lineplot = LinePlotView.init
            }

        subCmd : List (Cmd Msg)
        subCmd =
            [ MapView.init ] |> List.map Tuple.second

        cmd : Cmd Msg
        cmd =
            List.append
                [ Http.get
                    { url = "/schengen_data.csv"
                    , expect = Http.expectString DataReceived
                    }
                ]
                subCmd
                |> Cmd.batch

        model : Model
        model =
            ( global, local )
    in
    ( model, cmd )



-- UPDATE


update : Msg -> Model -> ( Model, Cmd Msg )
update msg ( global, local ) =
    case msg of
        ViewChange newView ->
            case newView of
                ForceDirectedGraph ->
                    let
                        fdg =
                            local.forcedirectedgraph

                        loadingfdg =
                            { fdg | loading = True }
                    in
                    ( ( { global | view = newView }, { local | forcedirectedgraph = loadingfdg } )
                    , ForceDirectedGraphView.calculateModel global.data loadingfdg |> Cmd.map ForceDirectedGraphView.propagate
                    )

                _ ->
                    ( ( { global | view = newView }, local ), Cmd.none )

        DataReceived (Ok rawData) ->
            case decode rawData of
                -- Pull the map data ONLY AFTER the Visa data was received.
                Ok data ->
                    ( ( { global | progress = Success, data = data }
                      , { local | forcedirectedgraph = ForceDirectedGraphView.firstInit }
                      )
                    , MapView.getMapData |> Cmd.map GotMap
                    )

                Err _ ->
                    ( ( { global | progress = Failure "Failed to parse data." }, local ), Cmd.none )

        DataReceived (Err error) ->
            ( ( { global | progress = buildErrorMessage error |> Failure }, local ), Cmd.none )

        GotMap subMsg ->
            let
                updateResult =
                    MapView.update subMsg global local.map
            in
            ( ( Triple.first updateResult, { local | map = Triple.second updateResult } ), Triple.third updateResult )

        GotDataTable subMsg ->
            let
                updateResult =
                    DataTableView.update subMsg global local.dataTable
            in
            ( ( Triple.first updateResult, { local | dataTable = Triple.second updateResult } ), Triple.third updateResult )

        GotForceDirectedGraph subMsg ->
            let
                updateResult =
                    ForceDirectedGraphView.update subMsg global local.forcedirectedgraph
            in
            ( ( Triple.first updateResult, { local | forcedirectedgraph = Triple.second updateResult } ), Triple.third updateResult )

        GotBarPlot subMsg ->
            let
                updateResult =
                    BarPlotView.update subMsg global local.barplot
            in
            ( ( Triple.first updateResult, { local | barplot = Triple.second updateResult } ), Triple.third updateResult )

        GotLinePlot subMsg ->
            let
                updateResult =
                    LinePlotView.update subMsg global local.lineplot
            in
            ( ( Triple.first updateResult, { local | lineplot = Triple.second updateResult } ), Triple.third updateResult )



-- VIEW

icon : Html Msg
icon =
    i [ Html.Attributes.class "fa-solid fa-house fa-xl" ] []

viewNav : Model -> Html Msg
viewNav ( global, _ ) =
    let
        activeClass : View -> String
        activeClass v =
            if global.view == v then
                "navbar-active"

            else
                "navbar-inactive"
    in
    nav [ Html.Attributes.class "navbar" ]
        [ button
            [ Html.Attributes.title "Home"
            , activeClass Home |> Html.Attributes.class, ViewChange Home |> onClick
            ] [ icon ]
        , div [ Html.Attributes.class "navbar-title" ]
            [ text "Schengen Visa Statistics" ]
        , div [ Html.Attributes.class "navbar-view-icons" ]
            [ button
                [ Html.Attributes.title "Map"
                , activeClass MapView |> Html.Attributes.class, ViewChange MapView |> onClick
                ] [ MapView.icon ]
            , button
                [ Html.Attributes.title "Force"
                , activeClass ForceDirectedGraph |> Html.Attributes.class, ViewChange ForceDirectedGraph |> onClick
                ] [ ForceDirectedGraphView.icon ]
            , button
                [ Html.Attributes.title "BarPlot"
                , activeClass BarPlot |> Html.Attributes.class, ViewChange BarPlot |> onClick
                ] [ BarPlotView.icon ]
            , button
                [ Html.Attributes.title "LinePlot"
                , activeClass LinePlot |> Html.Attributes.class, ViewChange LinePlot |> onClick
                ] [ LinePlotView.icon ]
            , button
                [ Html.Attributes.title "Table"
                , activeClass DataTable |> Html.Attributes.class, ViewChange DataTable |> onClick
                ] [ DataTableView.icon ]
            ]
        ]


viewHome : Global -> Html Msg
viewHome global =
    case global.progress of
        Loading ->
            Html.div [ Html.Attributes.style "margin" "auto" ]
                [ Html.div [ Html.Attributes.class "loader" ] []
                , Html.div
                    [ Html.Attributes.style "position" "relative"
                    , Html.Attributes.style "left" "1.5rem"
                    , Html.Attributes.style "bottom" "5rem"
                    ]
                    [ Html.text "Loading Data..." ]
                ]

        Failure msg ->
            Html.text (msg |> (++) "Error while loading data. ")

        Success ->
            div []
                [ Html.h2 []
                    [ text "Willkommen!" ]
                , Html.p []
                    [ text "Herzlich Willkommen zu unserem Projekt im Rahmen des Moduls "
                    , Html.b [] [ text "Informationsvisualisierung" ]
                    , text " im Sommersemester 2024 an der Martin-Luther-Universität Halle-Wittenberg!"
                    ]
                , Html.p []
                    [ text "In unserem Projekt geht es um die Aufbereitung und Visualisierung von Statistiken zu "
                    , Html.b [] [ text "Schengen Visa" ]
                    , text " für die Jahre 2014-2023."
                    , Html.br [] []
                    , text "Im Fokus steht dabei die Analyse der "
                    , Html.b [] [ text "Visafreundlichkeit" ]
                    , text " der Länder im Schengenraum, also die "
                    , Html.b [] [ text "a-priori Chance auf ein Visa" ]
                    , text " bei Antragstellung in der jeweiligen Auslandsvertretung (Konsulat)."
                    ]
                , Html.p []
                    [ text "Über die Icons in der Navigationsleiste kannst du zwischen den Visualisierungen wechseln." ]
                , Html.p []
                    [ text "Das Icon "
                    , MapView.icon
                    , text " führt dich zur "
                    , Html.b [] [ text "Weltkarte" ]
                    , text ", auf welcher du die Visafreundlichkeit der Schengenländer (grün) mittels "
                    , Html.b [] [ text "Chernoff Gesichtern" ]
                    , text " vergleichen kannst."
                    , text " Durch "
                    , Html.b [] [ text "Klick auf ein Drittland" ]
                    , text " (violett) und ggf. anschließende Auswahl eines "
                    , Html.b [] [ text "Konsulats" ]
                    , text " kannst du die zur Berechnung der Gesichter verwendeten Daten filtern."
                    ]
                , Html.p []
                    [ text "Mit dem Icon "
                    , ForceDirectedGraphView.icon
                    , text " kommst du zur "
                    , Html.b [] [ text "Force" ]
                    , text " Visualisierung der Beziehungen zwischen Schengen- und Drittländern."
                    , text " Auf der Legende kannst du ein "
                    , Html.b [] [ text "Schengenland selektieren" ]
                    , text ", sodass die zugehörigen Kanten in der Visualisierung rot markiert werden."
                    , text " Zusätzlich kann der Force-Graph mittels verschiedener Kräfte berechnet werden, was du durch "
                    , Html.b [] [ text "Klick der Version-Buttons" ]
                    , text " auswählen kannst."
                    , text " Die Größe der Schengenländer-Labels wird in der Visualisierung durch die"
                    , text " Beliebtheit, bzw. Anzahl erhaltener Visa-Anfragen, bestimmt."
                    , text " Durch "
                    , Html.b [] [ text "Auswahl eines Jahres" ]
                    , text " können die zur Berechnung zugrundegelegten Daten gefiltert werden."
                    , Html.br [] []
                    , Html.i [ Html.Attributes.style "font-size" "14px" ]
                        [ text "Bitte beachte, dass die Berechnung dieser Visualisierung bei unserer Datenmenge einige Zeit dauert." ]
                    ]
                , Html.p []
                    [ text "Mit dem Icon "
                    , BarPlotView.icon
                    , text " kommst du zum "
                    , Html.b [] [ text "BarPlot" ]
                    , text ", in welchem du die "
                    , Html.b [] [ text "Anzahl beantragter und bewilligter Visa pro Schengenland" ]
                    , text " approximativ ablesen kannst."
                    , text " Das Auswahlfeld ermöglicht dir einen "
                    , Html.b [] [ text "Wechsel zwischen bewilligt/beantragten Visa" ]
                    , text " im Diagramm."
                    , text " Außerdem kannst du eine "
                    , Html.b [] [ text "obere Grenze der Werte angeben" ]
                    , text " um ggf. kleinere Balken in den Fokus zu rücken."
                    ]
                , Html.p []
                    [ text "Mit dem Icon "
                    , LinePlotView.icon
                    , text " kommst du zum "
                    , Html.b [] [ text "LinePlot" ]
                    , text ", in welchem du eine "
                    , Html.b [] [ text "Zeitreihe der beantragten und bewilligten Visa" ]
                    , text " bzgl. eines Schengenlandes findest."
                    , text " Du kannst optional ein "
                    , Html.b [] [ text "Drittland auswählen" ]
                    , text " und die Einträge im Datensatz nach diesem filtern."
                    , text " Durch "
                    , Html.b [] [ text "Bewegung der Maus über die Grafik" ]
                    , text " kannst du die exakten Werte für ein konkretes Jahr ablesen."
                    ]
                , Html.p []
                    [ text "Letztlich wechselst du mit dem Icon "
                    , DataTableView.icon
                    , text " zur "
                    , Html.b [] [ text "Datentabelle" ]
                    , text ", in welcher du den von uns verwendeten Datensatz durchsuchen kannst."
                    , text " Durch "
                    , Html.b [] [ text "Eingabe in die Suchleiste" ]
                    , text " filterst du nach dem Namen des Schengenlandes, Drittlandes oder Konsulats."
                    , text " Außerdem kannst du den "
                    , Html.b [] [ text "Index eines Eintrags angeben" ]
                    , text " zu welchem du springen möchtest."
                    ]
                , Html.p []
                    [ text "Wir wünschen dir viel Spaß beim Ausprobieren und Analysieren!" ]
                , Html.i [ Html.Attributes.style "font-size" "14px" ]
                    [ text "Hinweis: In manchen Visualisierungen tauchen Flaggen Icons auf."
                    , text " Abhängig vom Browser und Betriebssystem werden diese ggf. nicht als Flaggen,"
                    , text " sondern mit zwei Buchstaben angezeigt."
                    ]
                ]


viewContent : Model -> Html Msg
viewContent ( global, local ) =
    case global.view of
        Home ->
            div [ Html.Attributes.class "content" ]
                [ viewHome global ]

        MapView ->
            div [ Html.Attributes.class "content" ]
                [ MapView.view global local.map ]

        DataTable ->
            div [ Html.Attributes.class "content" ]
                [ DataTableView.view global local.dataTable ]

        ForceDirectedGraph ->
            div [ Html.Attributes.class "content" ]
                [ ForceDirectedGraphView.view global local.forcedirectedgraph ]

        BarPlot ->
            div [ Html.Attributes.class "content" ]
                [ BarPlotView.view global local.barplot ]

        LinePlot ->
            div [ Html.Attributes.class "content" ]
                [ LinePlotView.view global local.lineplot ]


viewFooter : Model -> Html Msg
viewFooter model =
    div [ Html.Attributes.class "footer noselect" ]
        [ text "Erstellt von Aamon Hoffmann, Damian Haj-Houssin und Florian Raab mit "
        , a
            [ Html.Attributes.target "_"
            , Html.Attributes.href "https://elm-lang.org/docs"
            ]
            [ text "Elm" ]
        , text " und gebundlet mit "
        , a
            [ Html.Attributes.target "_"
            , Html.Attributes.href "https://parceljs.org/languages/elm/"
            ]
            [ text "Parcel" ]
        , text "."
        ]


view : Model -> Html Msg
view model =
    div [ Html.Attributes.class "wrapper" ]
        [ viewNav model
        , viewContent model
        , viewFooter model
        ]



-- SUBSCRIPTIONS


subscriptions : Model -> Sub Msg
subscriptions ( global, local ) =
    case global.view of
        DataTable ->
            DataTableView.subscriptions global local.dataTable

        ForceDirectedGraph ->
            ForceDirectedGraphView.subscriptions global local.forcedirectedgraph

        MapView ->
            MapView.subscriptions global local.map

        _ ->
            Sub.none
